@@ -8,16 +8,13 @@ use warnings;
 my $p_class = eval {
     require Net::IP::XS;
     return q{Net::IP::XS};
-} || eval {
-    require Net::IP;
-    return q{Net::IP};
 };
 
 if ( $p_class ) {
     $p_class->import;
 }
 else {
-    die "Both Net::IP and Net::IP::XS missing?\n";
+    die "Net::IP::XS is not loaded\n";
 }
 
 sub new {
@@ -70,37 +67,22 @@ sub version {
 }
 
 sub ip_is_ipv4 {
-    if ( $p_class eq 'Net::IP::XS' ) {
-        return Net::IP::XS::ip_is_ipv4( @_ );
-    }
-    else {
-        return Net::IP::ip_is_ipv4( @_ );
-    }
+    return Net::IP::XS::ip_is_ipv4( @_ );
 }
 
 sub ip_is_ipv6 {
-    if ( $p_class eq 'Net::IP::XS' ) {
-        return Net::IP::XS::ip_is_ipv6( @_ );
-    }
-    else {
-        return Net::IP::ip_is_ipv6( @_ );
-    }
+    return Net::IP::XS::ip_is_ipv6( @_ );
 }
 
 sub Error {
-    if ( $p_class eq 'Net::IP::XS' ) {
-        return Net::IP::XS::Error();
-    }
-    else {
-        return Net::IP::Error();
-    }
+    return Net::IP::XS::Error();
 }
 
 1;
 
 =head1 NAME
 
-Zonemaster::Engine::Net::IP - Net::IP/Net::IP::XS Wrapper (STILL EXPERIMENTAL)
+Zonemaster::Engine::Net::IP - Net::IP::XS Wrapper
 
 =head1 SYNOPSIS
 
