@@ -1,17 +1,18 @@
 package Zonemaster::Engine::Test::Delegation;
 
-use version; our $VERSION = version->declare("v1.0.16");
+use 5.014002;
 
 use strict;
 use warnings;
 
-use 5.014002;
-
-use Zonemaster::Engine;
+use version; our $VERSION = version->declare("v1.0.20");
 
 use List::MoreUtils qw[uniq];
 use Locale::TextDomain qw[Zonemaster-Engine];
 use Readonly;
+
+use Zonemaster::Engine::Profile;
+use Zonemaster::Engine::Recursor;
 use Zonemaster::Engine::Constants ':all';
 use Zonemaster::Engine::Net::IP;
 use Zonemaster::Engine::Test::Address;
@@ -66,6 +67,8 @@ sub metadata {
               NO_IPV4_NS_DEL
               NO_IPV6_NS_CHILD
               NO_IPV6_NS_DEL
+              TEST_CASE_END
+              TEST_CASE_START
               )
         ],
         delegation02 => [
@@ -76,12 +79,16 @@ sub metadata {
               DEL_NS_SAME_IP
               SAME_IP_ADDRESS
               DISTINCT_IP_ADDRESS
+              TEST_CASE_END
+              TEST_CASE_START
               )
         ],
         delegation03 => [
             qw(
               REFERRAL_SIZE_TOO_LARGE
               REFERRAL_SIZE_OK
+              TEST_CASE_END
+              TEST_CASE_START
               )
         ],
         delegation04 => [
@@ -90,14 +97,18 @@ sub metadata {
               IPV4_DISABLED
               IPV6_DISABLED
               ARE_AUTHORITATIVE
+              TEST_CASE_END
+              TEST_CASE_START
               )
         ],
         delegation05 => [
             qw(
               NO_NS_CNAME
-              NO_RESPONSE    
+              NO_RESPONSE
               NS_IS_CNAME
               UNEXPECTED_RCODE
+              TEST_CASE_END
+              TEST_CASE_START
               )
         ],
         delegation06 => [
@@ -106,6 +117,8 @@ sub metadata {
               SOA_EXISTS
               IPV4_DISABLED
               IPV6_DISABLED
+              TEST_CASE_END
+              TEST_CASE_START
               )
         ],
         delegation07 => [
@@ -114,15 +127,45 @@ sub metadata {
               EXTRA_NAME_CHILD
               TOTAL_NAME_MISMATCH
               NAMES_MATCH
+              TEST_CASE_END
+              TEST_CASE_START
               )
         ],
     };
 } ## end sub metadata
 
 Readonly my %TAG_DESCRIPTIONS => (
+    DELEGATION01 => sub {
+        __x    # DELEGATION:DELEGATION01
+          "Minimum number of name servers", @_;
+    },
+    DELEGATION02 => sub {
+        __x    # DELEGATION:DELEGATION02
+          "Name servers must have distinct IP addresses", @_;
+    },
+    DELEGATION03 => sub {
+        __x    # DELEGATION:DELEGATION03
+          "No truncation of referrals", @_;
+    },
+    DELEGATION04 => sub {
+        __x    # DELEGATION:DELEGATION04
+          "Name server is authoritative", @_;
+    },
+    DELEGATION05 => sub {
+        __x    # DELEGATION:DELEGATION05
+          "Name server must not point at CNAME alias", @_;
+    },
+    DELEGATION06 => sub {
+        __x    # DELEGATION:DELEGATION06
+          "Existence of SOA", @_;
+    },
+    DELEGATION07 => sub {
+        __x    # DELEGATION:DELEGATION07
+          "Parent glue name records present in child", @_;
+    },
     ARE_AUTHORITATIVE => sub {
         __x    # DELEGATION:ARE_AUTHORITATIVE
-          "All these nameservers are confirmed to be authoritative : {nsset}.", @_;
+          "All these nameservers are confirmed to be authoritative : {nsname_list}.", @_;
     },
     CHILD_DISTINCT_NS_IP => sub {
         __x    # DELEGATION:CHILD_DISTINCT_NS_IP
@@ -130,7 +173,7 @@ Readonly my %TAG_DESCRIPTIONS => (
     },
     CHILD_NS_SAME_IP => sub {
         __x    # DELEGATION:CHILD_NS_SAME_IP
-          "IP {address} in child refers to multiple nameservers ({nss}).", @_;
+          "IP {ns_ip} in child refers to multiple nameservers ({nsname_list}).", @_;
     },
     DEL_DISTINCT_NS_IP => sub {
         __x    # DELEGATION:DEL_DISTINCT_NS_IP
@@ -138,43 +181,43 @@ Readonly my %TAG_DESCRIPTIONS => (
     },
     DEL_NS_SAME_IP => sub {
         __x    # DELEGATION:DEL_NS_SAME_IP
-          "IP {address} in parent refers to multiple nameservers ({nss}).", @_;
+          "IP {ns_ip} in parent refers to multiple nameservers ({nsname_list}).", @_;
     },
     DISTINCT_IP_ADDRESS => sub {
         __x    # DELEGATION:DISTINCT_IP_ADDRESS
-          "All the IP addresses used by the nameservers are unique", @_;
+          "All the IP addresses used by the nameservers are unique.", @_;
     },
     ENOUGH_IPV4_NS_CHILD => sub {
         __x    # DELEGATION:ENOUGH_IPV4_NS_CHILD
-          "Child lists enough ({count}) nameservers ({nss}) "
-          . "that resolve to IPv4 addresses ({addrs}). Lower limit set to {minimum}.",
+          "Child lists enough ({count}) nameservers ({nsname_list}) "
+          . "that resolve to IPv4 addresses ({ns_ip_list}). Lower limit set to {minimum}.",
           @_;
     },
     ENOUGH_IPV4_NS_DEL => sub {
         __x    # DELEGATION:ENOUGH_IPV4_NS_DEL
-          "Delegation lists enough ({count}) nameservers ({nss}) "
-          . "that resolve to IPv4 addresses ({addrs}). Lower limit set to {minimum}.",
+          "Delegation lists enough ({count}) nameservers ({nsname_list}) "
+          . "that resolve to IPv4 addresses ({ns_ip_list}). Lower limit set to {minimum}.",
           @_;
     },
     ENOUGH_IPV6_NS_CHILD => sub {
         __x    # DELEGATION:ENOUGH_IPV6_NS_CHILD
-          "Child lists enough ({count}) nameservers ({nss}) "
-          . "that resolve to IPv6 addresses ({addrs}). Lower limit set to {minimum}.",
+          "Child lists enough ({count}) nameservers ({nsname_list}) "
+          . "that resolve to IPv6 addresses ({ns_ip_list}). Lower limit set to {minimum}.",
           @_;
     },
     ENOUGH_IPV6_NS_DEL => sub {
         __x    # DELEGATION:ENOUGH_IPV6_NS_DEL
-          "Delegation lists enough ({count}) nameservers ({nss}) "
-          . "that resolve to IPv6 addresses ({addrs}). Lower limit set to {minimum}.",
+          "Delegation lists enough ({count}) nameservers ({nsname_list}) "
+          . "that resolve to IPv6 addresses ({ns_ip_list}). Lower limit set to {minimum}.",
           @_;
     },
     ENOUGH_NS_CHILD => sub {
         __x    # DELEGATION:ENOUGH_NS_CHILD
-          "Child lists enough ({count}) nameservers ({nss}). Lower limit set to {minimum}.", @_;
+          "Child lists enough ({count}) nameservers ({nsname_list}). Lower limit set to {minimum}.", @_;
     },
     ENOUGH_NS_DEL => sub {
         __x    # DELEGATION:ENOUGH_NS_DEL
-          "Parent lists enough ({count}) nameservers ({glue}). Lower limit set to {minimum}.", @_;
+          "Parent lists enough ({count}) nameservers ({nsname_list}). Lower limit set to {minimum}.", @_;
     },
     EXTRA_NAME_CHILD => sub {
         __x    # DELEGATION:EXTRA_NAME_CHILD
@@ -186,11 +229,11 @@ Readonly my %TAG_DESCRIPTIONS => (
     },
     IPV4_DISABLED => sub {
         __x    # DELEGATION:IPV4_DISABLED
-          'IPv4 is disabled, not sending "{rrtype}" query to {ns}/{address}.', @_;
+          'IPv4 is disabled, not sending "{rrtype}" query to {ns}.', @_;
     },
     IPV6_DISABLED => sub {
         __x    # DELEGATION:IPV6_DISABLED
-          'IPv6 is disabled, not sending "{rrtype}" query to {ns}/{address}.', @_;
+          'IPv6 is disabled, not sending "{rrtype}" query to {ns}.', @_;
     },
     IS_NOT_AUTHORITATIVE => sub {
         __x    # DELEGATION:IS_NOT_AUTHORITATIVE
@@ -202,39 +245,39 @@ Readonly my %TAG_DESCRIPTIONS => (
     },
     NO_RESPONSE => sub {
         __x    # DELEGATION:NO_RESPONSE
-          "Nameserver {ns}/{address} did not respond.", @_;
+          "Nameserver {ns} did not respond.", @_;
     },
     NOT_ENOUGH_IPV4_NS_CHILD => sub {
         __x    # DELEGATION:NOT_ENOUGH_IPV4_NS_CHILD
-          "Child does not list enough ({count}) nameservers ({nss}) "
-          . "that resolve to IPv4 addresses ({addrs}). Lower limit set to {minimum}.",
+          "Child does not list enough ({count}) nameservers ({nsname_list}) "
+          . "that resolve to IPv4 addresses ({ns_ip_list}). Lower limit set to {minimum}.",
           @_;
     },
     NOT_ENOUGH_IPV4_NS_DEL => sub {
         __x    # DELEGATION:NOT_ENOUGH_IPV4_NS_DEL
-          "Delegation does not list enough ({count}) nameservers ({nss}) "
-          . "that resolve to IPv4 addresses ({addrs}). Lower limit set to {minimum}.",
+          "Delegation does not list enough ({count}) nameservers ({nsname_list}) "
+          . "that resolve to IPv4 addresses ({ns_ip_list}). Lower limit set to {minimum}.",
           @_;
     },
     NOT_ENOUGH_IPV6_NS_CHILD => sub {
         __x    # DELEGATION:NOT_ENOUGH_IPV6_NS_CHILD
-          "Child does not list enough ({count}) nameservers ({nss}) "
-          . "that resolve to IPv6 addresses ({addrs}). Lower limit set to {minimum}.",
+          "Child does not list enough ({count}) nameservers ({nsname_list}) "
+          . "that resolve to IPv6 addresses ({ns_ip_list}). Lower limit set to {minimum}.",
           @_;
     },
     NOT_ENOUGH_IPV6_NS_DEL => sub {
         __x    # DELEGATION:NOT_ENOUGH_IPV6_NS_DEL
-          "Delegation does not list enough ({count}) nameservers ({nss}) "
-          . "that resolve to IPv6 addresses ({addrs}). Lower limit set to {minimum}.",
+          "Delegation does not list enough ({count}) nameservers ({nsname_list}) "
+          . "that resolve to IPv6 addresses ({ns_ip_list}). Lower limit set to {minimum}.",
           @_;
     },
     NOT_ENOUGH_NS_CHILD => sub {
         __x    # DELEGATION:NOT_ENOUGH_NS_CHILD
-          "Child does not list enough ({count}) nameservers ({nss}). Lower limit set to {minimum}.", @_;
+          "Child does not list enough ({count}) nameservers ({nsname_list}). Lower limit set to {minimum}.", @_;
     },
     NOT_ENOUGH_NS_DEL => sub {
         __x    # DELEGATION:NOT_ENOUGH_NS_DEL
-          "Parent does not list enough ({count}) nameservers ({glue}). Lower limit set to {minimum}.", @_;
+          "Parent does not list enough ({count}) nameservers ({nsname_list}). Lower limit set to {minimum}.", @_;
     },
     NO_IPV4_NS_CHILD => sub {
         __x    # DELEGATION:NO_IPV4_NS_CHILD
@@ -262,11 +305,11 @@ Readonly my %TAG_DESCRIPTIONS => (
     },
     NS_IS_CNAME => sub {
         __x    # DELEGATION:NS_IS_CNAME
-          "Nameserver {ns} RR point to CNAME.", @_;
+          "Nameserver {nsname} RR points to CNAME.", @_;
     },
     NO_NS_CNAME => sub {
         __x    # DELEGATION:NO_NS_CNAME
-          "No nameserver point to CNAME alias.", @_;
+          "No nameserver points to CNAME alias.", @_;
     },
     REFERRAL_SIZE_TOO_LARGE => sub {
         __x    # DELEGATION:REFERRAL_SIZE_TOO_LARGE
@@ -278,7 +321,7 @@ Readonly my %TAG_DESCRIPTIONS => (
     },
     SAME_IP_ADDRESS => sub {
         __x    # DELEGATION:SAME_IP_ADDRESS
-          "IP {address} refers to multiple nameservers ({nss}).", @_;
+          "IP {ns_ip} refers to multiple nameservers ({nsname_list}).", @_;
     },
     SOA_EXISTS => sub {
         __x    # DELEGATION:SOA_EXISTS
@@ -286,7 +329,15 @@ Readonly my %TAG_DESCRIPTIONS => (
     },
     SOA_NOT_EXISTS => sub {
         __x    # DELEGATION:SOA_NOT_EXISTS
-          "A SOA query NOERROR response from {ns} was received empty.", @_;
+          "Empty NOERROR response to SOA query was received from {ns}.", @_;
+    },
+    TEST_CASE_END => sub {
+        __x    # DELEGATION:TEST_CASE_END
+          'TEST_CASE_END {testcase}.', @_;
+    },
+    TEST_CASE_START => sub {
+        __x    # DELEGATION:TEST_CASE_START
+          'TEST_CASE_START {testcase}.', @_;
     },
     TOTAL_NAME_MISMATCH => sub {
         __x    # DELEGATION:TOTAL_NAME_MISMATCH
@@ -294,7 +345,7 @@ Readonly my %TAG_DESCRIPTIONS => (
     },
     UNEXPECTED_RCODE => sub {
         __x    # DELEGATION:UNEXPECTED_RCODE
-          'Nameserver {ns}/{address} answered query with an unexpected rcode ({rcode}).', @_;
+          'Nameserver {ns} answered query with an unexpected rcode ({rcode}).', @_;
     },
 
 );
@@ -307,20 +358,49 @@ sub version {
     return "$Zonemaster::Engine::Test::Delegation::VERSION";
 }
 
+sub _ip_disabled_message {
+    my ( $results_array, $ns, @rrtypes ) = @_;
+
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $ns->address->version == $IP_VERSION_6 ) {
+        push @$results_array, map {
+          info(
+            IPV6_DISABLED => {
+                ns     => $ns->string,
+                rrtype => $_
+            }
+          )
+        } @rrtypes;
+        return 1;
+    }
+
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $ns->address->version == $IP_VERSION_4 ) {
+        push @$results_array, map {
+          info(
+            IPV4_DISABLED => {
+                ns     => $ns->string,
+                rrtype => $_,
+            }
+          )
+        } @rrtypes;
+        return 1;
+    }
+    return 0;
+}
+
 ###
 ### Tests
 ###
 
 sub delegation01 {
     my ( $class, $zone ) = @_;
-    my @results;
+    push my @results, info( TEST_CASE_START => { testcase => (split /::/, (caller(0))[3])[-1] } );
 
     # Determine delegation NS names
     my @del_nsnames = map { $_->string } @{ Zonemaster::Engine::TestMethods->method2( $zone ) };
     my $del_nsnames_args = {
-        count   => scalar( @del_nsnames ),
-        minimum => $MINIMUM_NUMBER_OF_NAMESERVERS,
-        glue    => join( q{;}, sort @del_nsnames ),
+        count       => scalar( @del_nsnames ),
+        minimum     => $MINIMUM_NUMBER_OF_NAMESERVERS,
+        nsname_list => join( q{;}, sort @del_nsnames ),
     };
 
     # Check delegation NS names
@@ -328,16 +408,15 @@ sub delegation01 {
         push @results, info( ENOUGH_NS_DEL => $del_nsnames_args );
     }
     else {
-        push @results,
-          info( NOT_ENOUGH_NS_DEL => $del_nsnames_args );
+        push @results, info( NOT_ENOUGH_NS_DEL => $del_nsnames_args );
     }
 
     # Determine child NS names
     my @child_nsnames = map { $_->string } @{ Zonemaster::Engine::TestMethods->method3( $zone ) };
     my $child_nsnames_args = {
-        count   => scalar( @child_nsnames ),
-        minimum => $MINIMUM_NUMBER_OF_NAMESERVERS,
-        nss     => join( q{;}, sort @child_nsnames ),
+        count       => scalar( @child_nsnames ),
+        minimum     => $MINIMUM_NUMBER_OF_NAMESERVERS,
+        nsname_list => join( q{;}, sort @child_nsnames ),
     };
 
     # Check child NS names
@@ -345,8 +424,7 @@ sub delegation01 {
         push @results, info( ENOUGH_NS_CHILD => $child_nsnames_args );
     }
     else {
-        push @results,
-          info( NOT_ENOUGH_NS_CHILD => $child_nsnames_args );
+        push @results, info( NOT_ENOUGH_NS_CHILD => $child_nsnames_args );
     }
 
     # Determine child NS names with addresses
@@ -357,16 +435,16 @@ sub delegation01 {
     my @child_ns_ipv6_addrs = uniq map { $_->address->short } grep { $_->address->version == 6 } @child_ns;
 
     my $child_ns_ipv4_args = {
-        count   => scalar( @child_ns_ipv4 ),
-        minimum => $MINIMUM_NUMBER_OF_NAMESERVERS,
-        nss     => join( q{;}, sort @child_ns_ipv4 ),
-        addrs   => join( q{;}, sort @child_ns_ipv4_addrs ),
+        count       => scalar( @child_ns_ipv4 ),
+        minimum     => $MINIMUM_NUMBER_OF_NAMESERVERS,
+        nsname_list => join( q{;}, sort @child_ns_ipv4 ),
+        ns_ip_list  => join( q{;}, sort @child_ns_ipv4_addrs ),
     };
     my $child_ns_ipv6_args = {
-        count   => scalar( @child_ns_ipv6 ),
-        minimum => $MINIMUM_NUMBER_OF_NAMESERVERS,
-        nss     => join( q{;}, sort @child_ns_ipv6 ),
-        addrs   => join( q{;}, sort @child_ns_ipv6_addrs ),
+        count       => scalar( @child_ns_ipv6 ),
+        minimum     => $MINIMUM_NUMBER_OF_NAMESERVERS,
+        nsname_list => join( q{;}, sort @child_ns_ipv6 ),
+        ns_ip_list  => join( q{;}, sort @child_ns_ipv6_addrs ),
     };
 
     if ( scalar( @child_ns_ipv4 ) >= $MINIMUM_NUMBER_OF_NAMESERVERS ) {
@@ -397,16 +475,16 @@ sub delegation01 {
     my @del_ns_ipv6_addrs = uniq map { $_->address->short } grep { $_->address->version == 6 } @del_ns;
 
     my $del_ns_ipv4_args = {
-        count   => scalar( @del_ns_ipv4 ),
-        minimum => $MINIMUM_NUMBER_OF_NAMESERVERS,
-        ns      => join( q{;}, sort @del_ns_ipv4 ),
-        addrs   => join( q{;}, sort @del_ns_ipv4_addrs ),
+        count       => scalar( @del_ns_ipv4 ),
+        minimum     => $MINIMUM_NUMBER_OF_NAMESERVERS,
+        nsname_list => join( q{;}, sort @del_ns_ipv4 ),
+        ns_ip_list  => join( q{;}, sort @del_ns_ipv4_addrs ),
     };
     my $del_ns_ipv6_args = {
-        count   => scalar( @del_ns_ipv6 ),
-        minimum => $MINIMUM_NUMBER_OF_NAMESERVERS,
-        ns      => join( q{;}, sort @del_ns_ipv6 ),
-        addrs   => join( q{;}, sort @del_ns_ipv6_addrs ),
+        count       => scalar( @del_ns_ipv6 ),
+        minimum     => $MINIMUM_NUMBER_OF_NAMESERVERS,
+        nsname_list => join( q{;}, sort @del_ns_ipv6 ),
+        ns_ip_list  => join( q{;}, sort @del_ns_ipv6_addrs ),
     };
 
     if ( scalar( @del_ns_ipv4 ) >= $MINIMUM_NUMBER_OF_NAMESERVERS ) {
@@ -429,14 +507,14 @@ sub delegation01 {
         push @results, info( NO_IPV6_NS_DEL => $del_ns_ipv6_args );
     }
 
-    return @results;
+    return ( @results, info( TEST_CASE_END => { testcase => (split /::/, (caller(0))[3])[-1] } ) );
 } ## end sub delegation01
 
 sub _find_dup_ns {
-    my %args = @_;
+    my %args          = @_;
     my $duplicate_tag = $args{duplicate_tag};
-    my $distinct_tag = $args{distinct_tag};
-    my @nss = @{ $args{nss} };
+    my $distinct_tag  = $args{distinct_tag};
+    my @nss           = @{ $args{ns_list} };
 
     my %nsnames_and_ip;
     my %ips;
@@ -456,8 +534,8 @@ sub _find_dup_ns {
             push @results,
               info(
                 $duplicate_tag => {
-                    nss     => join( q{;}, @{ $ips{$local_ip} } ),
-                    address => $local_ip,
+                    nsname_list => join( q{;}, @{ $ips{$local_ip} } ),
+                    ns_ip       => $local_ip,
                 }
               );
         }
@@ -472,7 +550,7 @@ sub _find_dup_ns {
 
 sub delegation02 {
     my ( $class, $zone ) = @_;
-    my @results;
+    push my @results, info( TEST_CASE_START => { testcase => (split /::/, (caller(0))[3])[-1] } );
 
     my @nss_del   = @{ Zonemaster::Engine::TestMethods->method4( $zone ) };
     my @nss_child = @{ Zonemaster::Engine::TestMethods->method5( $zone ) };
@@ -481,29 +559,29 @@ sub delegation02 {
       _find_dup_ns(
         duplicate_tag => 'DEL_NS_SAME_IP',
         distinct_tag  => 'DEL_DISTINCT_NS_IP',
-        nss           => [@nss_del],
+        ns_list       => [@nss_del],
       );
 
     push @results,
       _find_dup_ns(
         duplicate_tag => 'CHILD_NS_SAME_IP',
         distinct_tag  => 'CHILD_DISTINCT_NS_IP',
-        nss           => [@nss_child],
+        ns_list       => [@nss_child],
       );
 
     push @results,
       _find_dup_ns(
         duplicate_tag => 'SAME_IP_ADDRESS',
         distinct_tag  => 'DISTINCT_IP_ADDRESS',
-        nss           => [ @nss_del, @nss_child ],
+        ns_list       => [ @nss_del, @nss_child ],
       );
 
-    return @results;
+    return ( @results, info( TEST_CASE_END => { testcase => (split /::/, (caller(0))[3])[-1] } ) );
 } ## end sub delegation02
 
 sub delegation03 {
     my ( $class, $zone ) = @_;
-    my @results;
+    push my @results, info( TEST_CASE_START => { testcase => (split /::/, (caller(0))[3])[-1] } );
 
     my $long_name = _max_length_name_for( $zone->name );
     my @nsnames   = map { $_->string } @{ Zonemaster::Engine::TestMethods->method2( $zone ) };
@@ -519,14 +597,14 @@ sub delegation03 {
     }
 
     # If @nss_v4 is non-empty and all of its elements are in bailiwick of parent
-    if ( @nss_v4 && not grep { not $parent->name->is_in_bailiwick( $_->name ) } @nss_v4 ) {
+    if ( @nss_v4 and not grep { not $parent->name->is_in_bailiwick( $_->name ) } @nss_v4 ) {
         my $ns = $nss_v4[0];
         my $rr = Zonemaster::LDNS::RR->new( sprintf( q{%s IN A %s}, $ns->name, $ns->address->short ) );
         $p->unique_push( q{additional}, $rr );
     }
 
     # If @nss_v6 is non-empty and all of its elements are in bailiwick of parent
-    if ( @nss_v6 && not grep { not $parent->name->is_in_bailiwick( $_->name ) } @nss_v6 ) {
+    if ( @nss_v6 and not grep { not $parent->name->is_in_bailiwick( $_->name ) } @nss_v6 ) {
         my $ns = $nss_v6[0];
         my $rr = Zonemaster::LDNS::RR->new( sprintf( q{%s IN AAAA %s}, $ns->name, $ns->address->short ) );
         $p->unique_push( q{additional}, $rr );
@@ -550,12 +628,12 @@ sub delegation03 {
           );
     }
 
-    return @results;
+    return ( @results, info( TEST_CASE_END => { testcase => (split /::/, (caller(0))[3])[-1] } ) );
 } ## end sub delegation03
 
 sub delegation04 {
     my ( $class, $zone ) = @_;
-    my @results;
+    push my @results, info( TEST_CASE_START => { testcase => (split /::/, (caller(0))[3])[-1] } );
     my %nsnames;
     my @authoritatives;
     my $query_type = q{SOA};
@@ -564,27 +642,7 @@ sub delegation04 {
       my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
     {
 
-        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 ) {
-            push @results,
-              info(
-                IPV6_DISABLED => {
-                    ns      => $local_ns->name->string,
-                    address => $local_ns->address->short,
-                    rrtype  => $query_type,
-                }
-              );
-            next;
-        }
-
-        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 ) {
-            push @results,
-              info(
-                IPV4_DISABLED => {
-                    ns      => $local_ns->name->string,
-                    address => $local_ns->address->short,
-                    rrtype  => $query_type,
-                }
-              );
+        if ( _ip_disabled_message( \@results, $local_ns, $query_type ) ) {
             next;
         }
 
@@ -597,7 +655,7 @@ sub delegation04 {
                     push @results,
                       info(
                         IS_NOT_AUTHORITATIVE => {
-                            ns    => $local_ns->name->string,
+                            ns    => $local_ns->string,
                             proto => $usevc ? q{TCP} : q{UDP},
                         }
                       );
@@ -616,24 +674,24 @@ sub delegation04 {
                scalar @{ Zonemaster::Engine::TestMethods->method4( $zone ) }
             or scalar @{ Zonemaster::Engine::TestMethods->method5( $zone ) }
         )
-        and not scalar @results
+        and not grep { $_->tag ne q{TEST_CASE_START} } @results
         and scalar @authoritatives
       )
     {
         push @results,
           info(
             ARE_AUTHORITATIVE => {
-                nsset => join( q{,}, uniq sort @authoritatives ),
+                nsname_list => join( q{;}, uniq sort @authoritatives ),
             }
           );
     }
 
-    return @results;
+    return ( @results, info( TEST_CASE_END => { testcase => (split /::/, (caller(0))[3])[-1] } ) );
 } ## end sub delegation04
 
 sub delegation05 {
     my ( $class, $zone ) = @_;
-    my @results;
+    push my @results, info( TEST_CASE_START => { testcase => (split /::/, (caller(0))[3])[-1] } );
 
     my @nsnames = @{ Zonemaster::Engine::TestMethods->method2and3( $zone ) };
 
@@ -647,18 +705,11 @@ sub delegation05 {
             for my $key ( sort keys %nss ) {
                 my $ns = $nss{$key};
                 my $ns_args = {
-                    ns      => $ns->name->string,
-                    address => $ns->address->short,
-                    rrtype  => q{A},
+                    ns     => $ns->string,
+                    rrtype => q{A},
                 };
 
-                if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $ns->address->version == $IP_VERSION_6 ) {
-                    push @results, info( IPV6_DISABLED => $ns_args );
-                    next;
-                }
-
-                if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $ns->address->version == $IP_VERSION_4 ) {
-                    push @results, info( IPV4_DISABLED => $ns_args );
+                if ( _ip_disabled_message( \@results, $ns, q{A} ) ) {
                     next;
                 }
 
@@ -673,13 +724,13 @@ sub delegation05 {
                     next;
                 }
                 elsif ( scalar $p->get_records( q{CNAME}, q{answer} ) > 0 ) {
-                    push @results, info( NS_IS_CNAME => { ns => $local_nsname } );
+                    push @results, info( NS_IS_CNAME => { nsname => $local_nsname } );
                     next;
                 }
                 elsif ($p->is_redirect) {
                     my $p = $ns->query( $local_nsname, q{A}, { recurse => 1 } );
                     if ( defined $p and scalar $p->get_records( q{CNAME}, q{answer} ) > 0 ) {
-                        push @results, info( NS_IS_CNAME => { ns => $local_nsname } );
+                        push @results, info( NS_IS_CNAME => { nsname => $local_nsname } );
                     }
                 }
             }
@@ -687,7 +738,7 @@ sub delegation05 {
         else {
             my $p = Zonemaster::Engine::Recursor->recurse( $local_nsname, q{A} );
             if ( defined $p and scalar $p->get_records( q{CNAME}, q{answer} ) > 0 ) {
-                push @results, info( NS_IS_CNAME => { ns => $local_nsname } );
+                push @results, info( NS_IS_CNAME => { nsname => $local_nsname } );
             }
         }
     }
@@ -696,12 +747,12 @@ sub delegation05 {
         push @results, info( NO_NS_CNAME => {} );
     }
 
-    return @results;
+    return ( @results, info( TEST_CASE_END => { testcase => (split /::/, (caller(0))[3])[-1] } ) );
 } ## end sub delegation05
 
 sub delegation06 {
     my ( $class, $zone ) = @_;
-    my @results;
+    push my @results, info( TEST_CASE_START => { testcase => (split /::/, (caller(0))[3])[-1] } );
     my %nsnames;
     my $query_type = q{SOA};
 
@@ -709,27 +760,7 @@ sub delegation06 {
       my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
     {
 
-        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 ) {
-            push @results,
-              info(
-                IPV6_DISABLED => {
-                    ns      => $local_ns->name->string,
-                    address => $local_ns->address->short,
-                    rrtype  => $query_type,
-                }
-              );
-            next;
-        }
-
-        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 ) {
-            push @results,
-              info(
-                IPV4_DISABLED => {
-                    ns      => $local_ns->name->string,
-                    address => $local_ns->address->short,
-                    rrtype  => $query_type,
-                }
-              );
+        if ( _ip_disabled_message( \@results, $local_ns, $query_type ) ) {
             next;
         }
 
@@ -738,12 +769,7 @@ sub delegation06 {
         my $p = $local_ns->query( $zone->name, $query_type );
         if ( $p and $p->rcode eq q{NOERROR} ) {
             if ( not $p->get_records( $query_type, q{answer} ) ) {
-                push @results,
-                  info(
-                    SOA_NOT_EXISTS => {
-                        ns => $local_ns->name->string,
-                    }
-                  );
+                push @results, info( SOA_NOT_EXISTS => { ns => $local_ns->string } );
             }
         }
 
@@ -755,18 +781,18 @@ sub delegation06 {
                scalar @{ Zonemaster::Engine::TestMethods->method4( $zone ) }
             or scalar @{ Zonemaster::Engine::TestMethods->method5( $zone ) }
         )
-        and not scalar @results
+        and not grep { $_->tag ne q{TEST_CASE_START} } @results
       )
     {
         push @results, info( SOA_EXISTS => {} );
     }
 
-    return @results;
+    return ( @results, info( TEST_CASE_END => { testcase => (split /::/, (caller(0))[3])[-1] } ) );
 } ## end sub delegation06
 
 sub delegation07 {
     my ( $class, $zone ) = @_;
-    my @results;
+    push my @results, info( TEST_CASE_START => { testcase => (split /::/, (caller(0))[3])[-1] } );
 
     my %names;
     foreach my $name ( @{ Zonemaster::Engine::TestMethods->method2( $zone ) } ) {
@@ -817,7 +843,7 @@ sub delegation07 {
           );
     }
 
-    return @results;
+    return ( @results, info( TEST_CASE_END => { testcase => (split /::/, (caller(0))[3])[-1] } ) );
 } ## end sub delegation07
 
 ###
@@ -896,7 +922,7 @@ Verify that nameservers are authoritative.
 
 =item delegation05($zone)
 
-Verify that NS RRs do not point to CNAME alias.
+Verify that NS RRs do not points to CNAME alias.
 
 =item delegation06($zone)
 
